# Calculating image and channel statistics
# Ver. 0.7a Interactive Dialog Version (under developping)
# (c)2021-22 Ohnishi, Yasuo / All right reserved.

from ij import IJ, ImagePlus, CompositeImage
from ij import WindowManager as WM
from ij.io import DirectoryChooser, OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.plugin import Duplicator
from ij.process import ImageConverter, ImageProcessor
from loci.plugins import BF
import math
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')
import datetime
from java.awt.event import ItemListener, TextListener
from java.lang import Integer

#---Convert Value in 8, 16, 32bit mode
def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = Value * 256
    elif Mode == "32-bit":
        Value = Value * 256 * 256
    else:
        Value = Value
    return Value
#
def convValue2(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = float(Value) / 256.0
    elif Mode == "32-bit":
        Value = float(Value) / 256.0 / 256.0
    return Value

#------ calculating B-G mean difference routine ############

def calcMean(imp_0_org, imp_1_org, imp_Gray_org, xStart, xEnd, \
                nYDivide, yStart, yEnd, L_thsld_lw, L_thsld_up, Mode):
#   ip0: B image / ip1: G image / xStart & xEnd: Starting & Ending Point (x) of ROI
#   nYDivide:  number of dividing zones in Y
    imp_0 = Duplicator().run(imp_0_org)
    imp_1 = Duplicator().run(imp_1_org)
    imp_Gray = Duplicator().run(imp_Gray_org)
    print ("Mean calculate routine xStart - xEnd", xStart, xEnd, \
                                            "yStart - yEnd", yStart, yEnd)
#   Atteintion! imp_Gray is always 8bit!!
    if int(L_thsld_up) -  int(L_thsld_lw) > 0: 
                # calculation will be limited within luminance range or not
        L_thsld_up = int(L_thsld_up)
        L_thsld_lw = int(L_thsld_lw)
    else:
        L_thsld_up = 256 - 1 # if luminance threshold is 0 calculate all range
    print "for debug luminance range: ", L_thsld_lw, L_thsld_up
    impTmp = []
    impTmp.append(imp_0)
    impTmp.append(imp_1)
    width = imp_0.getWidth() # get image width
    height = imp_0.getHeight() # get image height
    pixelNum = width * height # get total pixel number
    if yStart == 0 and yEnd == 0:
        yEnd = height
#    else:
#        height2 = yEnd - yStart
    maxDiff = 0
    minDiff = 0
    v_x = []
    v_y = []
    v_0 = []
    v_1 = []
    v_gray = []
    v_diff = []
    n = 0 # for roop counter
    for x in range(xStart, xEnd): # processing in each X row
        for y in range(yStart, yEnd): 
            PxGray = imp_Gray.getPixel(x, y) # get Second Imp pixel Value
            if PxGray[0] <= L_thsld_up and PxGray[0] >= L_thsld_lw: 
                # Attention! if you use getPixel in imagePlus, 
                # the got value will be array!
                v_x.append(x)
                v_y.append(y)
                tmpv0 = impTmp[0].getPixel(x, y)
                tmpv1 = impTmp[1].getPixel(x, y)
                v_0.append(tmpv0[0]) # get first Imp pixel Value
                v_1.append(tmpv1[0]) # get second Imp pixel Value
                v_gray.append(imp_Gray.getPixel(x, y)) 
                        # get grayscale Imp pixel Value
                v_diff.append(tmpv0[0] - tmpv1[0]) 
                        # get grayscale Imp pixel Value
                n = n + 1 # counter 1 plus
#       ----   
    print "for debug counted pixels: ", n
    mean = sum(v_diff) / float(len(v_diff))
    mean_imp0 =   sum(v_0) / float(len(v_0))
    mean_imp1 =   sum(v_1) / float(len(v_1))
    maxDiff = max(v_diff)
    minDiff =  min(v_diff)
    impTmp[0].close()
    impTmp[1].close()

    print "mean", convValue2(mean, Mode), "mean imp0", \
        convValue2(mean_imp0, Mode), "mean imp1", convValue2(mean_imp1, Mode)
    sum_for_v = 0
    sum_for_v_imp0 = 0
    sum_for_v_imp1 = 0
    for i in range(len(v_diff)):
        sum_for_v += (v_diff[i] - mean) ** 2
        sum_for_v_imp0 += (v_0[i] - mean_imp0) ** 2
        sum_for_v_imp1 += (v_1[i] - mean_imp1) ** 2
    v = sum_for_v / float(len(v_diff)) #variance
    v_imp0 = sum_for_v_imp0 / float(len(v_0))
    v_imp1 = sum_for_v_imp1 / float(len(v_1))
    # co-variance
    sum_imp0imp1 = 0
    for i, j in zip(v_0, v_1):
        sum_imp0imp1 += (i - mean_imp0) * (j - mean_imp1)
        cov = sum_imp0imp1 / len(v_0)

    sd = math.sqrt(v) # standard deviation
    sd_imp0 = math.sqrt(v_imp0)
    sd_imp1 = math.sqrt(v_imp1)
    if sd_imp0 * sd_imp1 != 0:
        peason_r = cov / (sd_imp0 * sd_imp1) # Peason R
    else:
        peason_r = 999
    print "Calculation Luminance Range:", L_thsld_lw, " - ", L_thsld_up
    print "Diff variance", convValue2(v, Mode), "sd", convValue2(sd, Mode), \
        "Max:", convValue2(maxDiff, Mode), "Min:", convValue2(minDiff, Mode)
    print "imp0 variance", convValue2(v_imp0, Mode), "sd", \
                                        convValue2(sd_imp0, Mode)
    print "imp1 variance", convValue2(v_imp1, Mode), "sd", \
                                        convValue2(sd_imp1, Mode)
    print "covariance", convValue2(cov, Mode), "peason r", peason_r

    imp_0.close()
    imp_1.close()
    imp_Gray.close()
#       ----   
    return convValue2(mean, Mode), convValue2(mean_imp0, Mode), \
        convValue2(mean_imp1, Mode), convValue2(v, Mode), \
        convValue2(sd, Mode), convValue2(maxDiff, Mode), \
        convValue2(minDiff, Mode), convValue2(v_imp0, Mode), \
        convValue2(sd_imp0, Mode), convValue2(v_imp1, Mode), \
        convValue2(sd_imp1, Mode), convValue2(cov, Mode), peason_r
           # Return value: mean B-G difference 

#-------------------------
def write_out(mean, mean_imp0, mean_imp1, v, sd, maxDiff, minDiff, \
    v_imp0, sd_imp0, v_imp1, sd_imp1, cov, peason_r, L_thsld_lw, \
    L_thsld_up2, xStart, xEnd, yStart, yEnd, color1, color2, result_file):
    line = "\n1番目の色Ch: " + color1 + " 2番目の色Ch: " +  color2
    result_file.write(line)
    line = "\n計算対象輝度範囲(グレースケール): " + str(L_thsld_lw) + \
                                                " - " + str(L_thsld_up2)
    result_file.write(line)
    line = "\n計算対象ROI: X: " + str(xStart) + " - " + str(xEnd) + \
                                " /Y: " + str(yStart) + " - " + str(yEnd) 
    result_file.write(line)
    line = "\n色Ch間ピクセル輝度差の平均: {0:.2f} Ch 1平均輝度: {1:.2f}"\
        .format(mean, mean_imp0) + " Ch 2平均輝度: {0:.2f}".format(mean_imp1)
    result_file.write(line)
    line = "\n輝度差分散: {0:.2f} 輝度差標準偏差: {1:.2f}".format(v, sd) + \
        " 輝度最高差: {0:.2f} 輝度最低差: {1:.2f}".format(maxDiff, minDiff)
    result_file.write(line)
    line =  "\nCh1分散: {0:.2f} Ch1標準偏差: {1:.2f}"\
        .format(v_imp0, sd_imp0) + " Ch2分散: {0:.2f} Ch2標準偏差: {1:.2f}"\
                                                .format(v_imp1, sd_imp1)
    result_file.write(line)
    line = "\nCh1-2 共分散: {0:.2f} ピアソン相関係数: {1:.2f}"\
                                            .format(cov, peason_r)
    result_file.write(line)
    if peason_r < 0.9:
        line = "\nマニュアル補正調整量目安: {0:.2f}".format(sd - abs(mean))
        result_file.write(line)
    result_file.write('\n--------------------------------\n')

#-------------------------

def manipulation(imp0, firstImp, secondImp, Mode, Show, xStart, \
    xEnd, folder, filename, yStart, yEnd, calcAll, L_thsld_lw, L_thsld_up):
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    n = 0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
        ip = i.getProcessor()
        n = n + 1
    impR =  impsCh[0]
    impG =  impsCh[1]
    impB =  impsCh[2]
    """
    if Mode <> "8bit":
        IJ.saveAsTiff(impR, folder + filename +"_R.tif")
        IJ.saveAsTiff(impG, folder + filename +"_G.tif")
    """
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
    width = impB.getWidth()
    print ("xEnd - xStart:", xEnd - xStart)
    if xEnd - xStart >= 1:
        mergeimp = RGBStackMerge.mergeChannels([impR, impG, impB, \
                                        None, None, None, None], True)
        RGBStackConverter.convertToRGB(mergeimp)
        if Mode == "8-bit":
            ImageConverter(mergeimp).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(mergeimp).convertToGray16()
        else:
            ImageConverter(mergeimp).convertToGray32()

        if firstImp == "Red":
            imp_0 = impR
        elif firstImp == "Green":
            imp_0 = impG
        elif firstImp == "Blue":
            imp_0 = impB
        elif firstImp == "Gray":
            imp_0 = mergeimp

        if secondImp == "Red":
            imp_1 = impR
        elif secondImp == "Green":
            imp_1 = impG
        elif secondImp == "Blue":
            imp_1 = impB
        elif secondImp == "Gray":
            imp_1 = mergeimp
        print ("execute calcMean")
        print ("--------------------------------")

        if L_thsld_up == 0.0: # set L_thsld_up for printing
            L_thsld_up2 = 255.9
        else: 
            L_thsld_up2 = L_thsld_up

        result_file = open(folder + filename + "_result" + \
            str(int(L_thsld_lw)) + "-" + str(int(L_thsld_up2)) + ".txt", 'w')
        result_file.write('--------------------------------\n')
        result_file.write(u'画像ファイル統計: '+ filename)


        if calcAll == False:
            print ("1st Imp:", firstImp, " 2nd Imp:", secondImp)
            mean, mean_imp0, mean_imp1, v, sd, maxDiff, minDiff, \
                v_imp0, sd_imp0, v_imp1,  sd_imp1, cov, peason_r = \
            calcMean(imp_0, imp_1, mergeimp, xStart, xEnd, 1, yStart, \
                yEnd, L_thsld_lw, L_thsld_up, Mode)
            write_out(mean, mean_imp0, mean_imp1, v, sd, maxDiff, \
                minDiff, v_imp0, sd_imp0, v_imp1, sd_imp1, cov, peason_r, \
                L_thsld_lw, L_thsld_up2, xStart, xEnd, yStart,\
                                    yEnd, firstImp, secondImp, result_file)
        else:
#           R & G
            print ("1st Imp: Red"," 2nd Imp: Green")
            mean, mean_imp0, mean_imp1, v, sd, maxDiff, minDiff, \
                v_imp0, sd_imp0, v_imp1, sd_imp1, cov, peason_r = \
                calcMean(impR, impG, mergeimp, xStart, xEnd, 1,\
                             yStart, yEnd, L_thsld_lw, L_thsld_up, Mode)
            write_out(mean, mean_imp0, mean_imp1, v, sd, maxDiff, \
                minDiff, v_imp0, sd_imp0, v_imp1, sd_imp1, cov, peason_r, \
                L_thsld_lw, L_thsld_up2, xStart, xEnd, yStart,\
                                         yEnd, "Red", "Green", result_file)
#           R & B
            print ("--------------------------------")
            print ("1st Imp: Red"," 2nd Imp: Blue")
            mean, mean_imp0, mean_imp1, v, sd, maxDiff, minDiff, v_imp0, \
                sd_imp0, v_imp1, sd_imp1, cov, peason_r = \
                calcMean(impR, impB, mergeimp, xStart, xEnd, 1, \
                               yStart, yEnd, L_thsld_lw, L_thsld_up, Mode)
            write_out(mean, mean_imp0, mean_imp1, v, sd, maxDiff, \
                minDiff, v_imp0, sd_imp0,  v_imp1, sd_imp1, cov, peason_r, \
                L_thsld_lw, L_thsld_up2, xStart, xEnd, yStart, yEnd, "Red", \
                                                        "Blue", result_file)
#           G & B
            print ("--------------------------------")
            print ("1st Imp: Green"," 2nd Imp: Blue")
            mean, mean_imp0, mean_imp1, v, sd, maxDiff, minDiff, v_imp0, \
                sd_imp0, v_imp1, sd_imp1, cov, peason_r = \
                calcMean(impG, impB, mergeimp, xStart, xEnd, 1, yStart, \
                                        yEnd, L_thsld_lw, L_thsld_up, Mode)

            write_out(mean, mean_imp0, mean_imp1, v, sd, maxDiff, minDiff,\
                v_imp0, sd_imp0,  v_imp1, sd_imp1, cov, peason_r, \
                L_thsld_lw, L_thsld_up2, xStart, xEnd, yStart, yEnd, \
                                            "Green",  "Blue", result_file)
        result_file.close()
    """
    if Mode <> "8-bit":
        IJ.saveAsTiff(impB2, folder + filename +"_B2.tif")
    """

    if Show == "Yes":
        imp0.hide()
        impR.hide()
        impG.hide()
        impB.hide()
#        mergeimp.hide()
    imp0.close()
    impR.close()
    impG.close()
    impB.close()
#    mergeimp.close()

# ---ROI dialog-------
class RoiPreviewer(ItemListener, TextListener):
    def __init__(self, imp, ckbox1, nfield_xStart, nfield_xEnd, \
                                                nfield_yStart, nfield_yEnd):
        """
             imp: an ImagePlus
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                        dynamically update the ImagePlus as the
                                        scrollbar is updated, or not.
        """
        self.imp = imp
        self.original_ip = imp.getProcessor().duplicate() # store a copy
        self.nfield_xStart = nfield_xStart
        self.nfield_xEnd = nfield_xEnd
        self.nfield_yStart = nfield_yStart
        self.nfield_yEnd = nfield_yEnd
        self.prv_ckBox1 = ckbox1
    

    def textValueChanged(self, event):
        if self.prv_ckBox1.getState() == True:
            self.changeRoi()

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data 
                        on what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.changeRoi()
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)
    
    def changeRoi(self):
        """ Execute the in-place scaling of the ImagePlus. """
        rx = Integer.parseInt(self.nfield_xStart.getText())
        ry = Integer.parseInt(self.nfield_yStart.getText())
        rxe = Integer.parseInt(self.nfield_xEnd.getText())
        rye = Integer.parseInt(self.nfield_yEnd.getText())
        rw = rxe - rx
        rh = rye - ry
        rx = int(rx)
        ry = int(ry)
        rw = int(rw)
        rh = int(rh)
        if rx + rw > self.imp.width:
           rw = self.imp.width - rx
        if ry + rh > self.imp.height:
           rh = self.imp.height - ry
        print (rx, ry, rw, rh)
        self.imp.setRoi(rx, ry, rw, rh)

def setRoiUI(imp, Title, Message, yFixed, nfield_yStart, nfield_yEnd, Mode):
#    imp = WM.getCurrentImage()
    impw = imp.width
    imph = imp.height
    roi = imp.getRoi()
    if roi:
        rOrgx = roi.getBounds().x
        rOrgy = roi.getBounds().y
        rOrgw = roi.getBounds().width
        rOrgh = roi.getBounds().height
    else:
        rOrgx = 0
        rOrgy = 0
        rOrgw = impw
        rOrgh = imph
    
    if not imp:
        print "Open an image first!"
        Mode = "Stop"
        return
    gd = GenericDialog(Title)
    gd.addMessage(Message)
    gd.addNumericField("Start Point (x)", rOrgx, 0)
    gd.addNumericField("End Point (x)", rOrgx + rOrgw, 0)
    if yFixed == "No":
        gd.addMessage("Y Region to be corrected")
        gd.addNumericField("Start Point (y)", rOrgy, 0)
        gd.addNumericField("End Point (y)", rOrgy + rOrgh, 0)
    else:
        gd.addMessage("Y Region to be sampled")
        gd.addMessage("Start Point (y):" + nfield_yStart.getText())
        gd.addMessage("End Point (y):" + nfield_yEnd.getText())
    gd.addCheckbox("Preview ROI", True)

    # The UI elements for the above two inputs
    nfield_xStart = gd.getNumericFields().get(0) # 1st start
    nfield_xEnd = gd.getNumericFields().get(1) # 1st end
    if yFixed == "No":
        nfield_yStart = gd.getNumericFields().get(2) # 1st start
        nfield_yEnd = gd.getNumericFields().get(3) # 1st end

    checkbox1 = gd.getCheckboxes().get(0)


    previewer = RoiPreviewer(imp, checkbox1, nfield_xStart, nfield_xEnd, \
                                                nfield_yStart, nfield_yEnd)
    checkbox1.addItemListener(previewer)
    nfield_xStart.addTextListener(previewer)
    nfield_xEnd.addTextListener(previewer)
    nfield_yStart.addTextListener(previewer)
    nfield_yEnd.addTextListener(previewer)

    gd.showDialog()

    if gd.wasCanceled():
        previewer.reset()
        Mode   = "Stop"
        IJ.log("User Canceled")
        checkbox1.removeItemListener(previewer)
    else:
        previewer.changeRoi()
        checkbox1.removeItemListener(previewer)
    return nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd, Mode

#--parameter dialog 
### Get parameters from the input DialogBox 
def getParameters(Mode):
    gd = GenericDialog("Image channel statistics: (c)2022 Yasuo Ohnishi")
    gd.addMessage("Please Enter Parameters")
    gd.addCheckbox("Resize to width 1280", True)
    gd.addCheckbox("Calculate all channels", False)
    FirstImpSelect = ["Red","Green","Blue","Gray"]
    gd.addChoice("Select First Image", FirstImpSelect, FirstImpSelect[2])
    SecondImpSelect = ["Red","Green","Blue","Gray"]
    gd.addChoice("Select Second Image", SecondImpSelect, SecondImpSelect[1])
    Show = ["Yes","No"]
    gd.addMessage("If you would like to limit calculation luminance range " +\
                                                "\n Please Enter Parameters")
    gd.addNumericField("Lower Luminance threshold", 0, 0)
    gd.addNumericField("Upper Luminance threshold", 0, 0)
    gd.addChoice("Display working windows", Show, Show[0])
    correction = ["Yes","No"]

    gd.showDialog()  # Show Dialog

#
    Resize = gd.getNextBoolean()
    calcAll = gd.getNextBoolean()
    firstImp   = gd.getNextChoice()
    secondImp   = gd.getNextChoice()
    L_thsld_lw = gd.getNextNumber()
    L_thsld_up = gd.getNextNumber()
    parm_Show   = gd.getNextChoice()
#   ---- end of parameter input 
#
    if gd.wasCanceled():  
        Mode  = "Stop"
        IJ.log("Canceled")
#
    return Mode, str(parm_Show), str(firstImp), str(secondImp), calcAll, \
                                            Resize, L_thsld_lw, L_thsld_up

#---main--------
meanIntensity = []
window = WM.getActiveWindow()
imgCurrent = WM.getCurrentImage()
Mode = "Auto"
if not imgCurrent:
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    filename = od.getFileName()
    if filename:
        path = folder + filename
        IJ.log(path)
        IJ.log(folder)
        IJ.log(filename)
# Always Open file with Bio-format Plugin
        imps = BF.openImagePlus(folder + filename) 
                # Open file with Bio-format Plugin
        imp0 = imps[0] #imp0: Original Image 
        imp0.show()
    else:
        Mode ="Stop"
else:
    imp0 = imgCurrent
    filename = imp0.getTitle()
    try:
        folder = imp0.getOriginalFileInfo().directory
    except:
        folder = IJ.getDirectory("Output_directory")
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
        Mode = "8-bit"
    elif bitdepth ==16:
        Mode = "16-bit"
    elif bitdepth ==32:
        Mode = "32-bit"
    else:
        Mode = "16-bit"
Mode = str(Mode)
if Mode != "Stop":
    Mode, Show, firstImp, secondImp, calcAll, Resize, L_thsld_lw, \
                                    L_thsld_up = getParameters(Mode)
    print ("Resize:", Resize)
if Resize == True:
    scale = 1280.0 / float(imp0.width)
    new_height = float(imp0.height) * scale # preview image height
    print ("height", imp0.height)
    new_height = int(new_height)
    print ("new height", new_height)
    tmp_imp0 = imp0.resize(1280, new_height, "none")
    imp0.close()
    imp0 = tmp_imp0
    imp0.show()
    imp0.updateAndDraw()
    WM.getActiveWindow().toFront()
    print ("execute resize:", imp0.width)
if Mode != "Stop":
    Title = "Image channel statistics"
    Message = "Select calculating ROI"
    yFixed = "No"
    nfield_xStart, nfield_xEnd, nfield_yStart, nfield_yEnd, Mode = \
        setRoiUI(imp0, Title, Message, yFixed, 0, 0, Mode)
    WM.getActiveWindow().toFront()
if Mode != "Stop":
    xStart = Integer.parseInt(nfield_xStart.getText())
    xEnd = Integer.parseInt(nfield_xEnd.getText())
    yStart = Integer.parseInt(nfield_yStart.getText())
    yEnd = Integer.parseInt(nfield_yEnd.getText())
    if xEnd > imp0.width:
        xEnd =  imp0.width
    if yEnd > imp0.height:
        yEnd =  imp0.height
if Mode != "Stop":
    manipulation(imp0, firstImp, secondImp, Mode, Show, xStart, xEnd, \
        folder, filename, yStart, yEnd, calcAll, L_thsld_lw, L_thsld_up)
imp0.hide()
imp0.close()
window = WM.getActiveWindow()
if window:
    WM.closeAllWindows()
    #WM.removeWindow()

IJ.log("Finish")
#IJ.run("Quit")

